import os
import pandas as pd
import numpy as np
import xlrd
import openpyxl
import json

def process_json(filepath):
    #extract report text from json file
    with open(filepath) as f:
        full_json = json.load(f)
    report = full_json['Report']
    return [report]

def process_spreadsheet(filepath):
    _, input_extension = os.path.splitext(filepath)
    if input_extension == '.csv':
        df = pd.read_csv(filepath)
    elif input_extension == '.xlsx':
        df = pd.read_excel(filepath)
    reports = df['Report']
    return reports
    
def aggregate_reports(report_list, save_dir):
    #bundle all reports as numpy array for de-id algorithm
    np_reports = np.array(report_list)
    save_path = os.path.join(save_dir, 'reports.npy')
    np.save(save_path, np_reports) 
    
def parse_input(input_filepath, scratch_dir, logger):
    try:
        _, input_extension = os.path.splitext(input_filepath)
        logger.info(f'Reformatting input file of type {input_extension}')
        if input_extension == '.csv' or input_extension == '.xlsx':
            reports = process_spreadsheet(input_filepath)
        elif input_extension == '.json':
            reports = process_json(input_filepath) 
        else:
            logger.error(f'Files of type {ext} are not accepted as input.')
            raise ValueError(f'Files of type {ext} are not accepted as input.')
    except Exception as e:
        logger.error(f'The following exception was encountered while reformatting input: {e}')
    aggregate_reports(reports, scratch_dir)
    logger.info(f'Successfully reformatted {len(reports)} reports.')


        