import numpy as np
import os
import json
import pandas as pd
import time

def process_spreadsheet(output_dir, 
                        input_filepath, 
                        input_extension, 
                        deid_reports):
    input_filename = os.path.basename(input_filepath)
    if input_extension == '.csv':
        df = pd.read_csv(os.path.join(input_filepath))
    elif input_extension == '.xlsx':
        df = pd.read_excel(os.path.join(input_filepath))
    df['Report'] = deid_reports
    df = df.rename(columns={'Facility ID': 'Facility ID',
                       'Patient ID': 'Patient ID',
                       'Accession Number': 'Accession Number',
                       'Report': 'De-ID Report'})
    if input_extension == '.csv':
        output_filename = input_filename.split('.csv')[0] + '_deidentified_' + str(time.time()) + '.csv'
        df.to_csv(os.path.join(output_dir, output_filename), index=False)
    if input_extension == '.xlsx':
        output_filename = input_filename.split('.xlsx')[0] + '_deidentified_' + str(time.time()) + '.xlsx'
        df.to_excel(os.path.join(output_dir, output_filename), index=False)

def process_json(output_dir,
                 input_filepath,
                 deid_report):
    input_filename = os.path.basename(input_filepath)
    with open(input_filepath) as f:
        full_json = json.load(f)
    full_json['Report'] = deid_report
    full_json = {'De-ID Report' if k == 'Report' else k:v for k,v in full_json.items()}
    output_filename = input_filename.split('.json')[0] + '_deidentified_' + str(time.time()) + '.json'
    with open(os.path.join(output_dir, output_filename), 'w') as f:
        json.dump(full_json, f)

def parse_output(output_dir,
                 input_filepath,
                 model_output_path,
                 logger):
    try:
        _, input_extension = os.path.splitext(input_filepath)
        logger.info('Processeing de-id results into file of type {input_extention}')
        with open(model_output_path, "rb") as f:
            deid_reports = np.load(f, allow_pickle=True)
        if input_extension == '.csv' or input_extension == '.xlsx':
            process_spreadsheet(output_dir, input_filepath, input_extension, deid_reports)
        elif input_extension == '.json':
            if len(deid_reports) != 1:
                deid_report = '\n'.join(deid_reports)
            else:
                deid_report = deid_reports[0]
            process_json(output_dir, input_filepath, deid_report)
        logger.info(f'Output file saved successfully.')
    except Exception as e:
        logger.error(f'The following exception was encountered while reformatting output: {e}')

    

              

